"""
A simple to use Python Library used to handle API Calls for all kinds of unique discord bots.

:copyright: (c) 2021-present HiddenStorm
:license: MIT, see LICENSE for more details.
"""

__name__ = "DiscAPI"
__version__ = "0.2.0a"
__author__ = "HiddenStorm"

from .Client import Client
from .Guild import Guild



import asyncio
import time
import threading

from .HTTPSCLIENT import HTTPClient, op
from .Log import Log
from .Guild import Guild

class Client:
    def __init__(self, Log = False):
        """
        The Client class to instantiate a Discord Client.
        
        Parameters
        ----------

        Optional: Bool Log | Default = False
            Prints out extra information.
        """

        self.Log = Log
        self.Loop = asyncio.get_event_loop()
        self.handlers = {}
        self.uptime = 0
        self.guilds = []

    def run(self, TOKEN, SECRET):
        """
        This function starts the client.
        Return nothing.

        Parameters
        ----------

        Required: String Token
            The Client Token you get on https://discord.com/developers/applications/{AppID}/bot
        
        Required: String Secret
            The Client Secret you get on https://discord.com/developers/applications/{AppID}/oauth2
        """

        self.Token=TOKEN
        self.Secret=SECRET
        self.HTTPClient = HTTPClient(self)
        self.HTTPClient.connect()
        self.UptimeThread = threading.Thread(target=self.counttime)
        self.UptimeThread.start()
        asyncio.run(self.call_event("on_ready"))
        self.Loop.run_forever()
        
    def event(self, method):
        """
        Putting
        @client.event("event")
        above a function will make it get called everytime that specific event happens.

        Events
        ------

        on_ready(): gets called when the client is ready to interact with discord.

        on_message(): gets called when a message gets posted into a discord guild and includes a message object in the function parameters.

        Example
        -------

        >>> @client.event("on_ready")
        >>> async def my_function():
        >>>     print("Bot is logged in.")
        """
        
        Log.Debug(self, "Registered " + method + " as an event.")
        def registerhandler(handler):
            if method in self.handlers:
                self.handlers[method].append(handler)
            else:
                self.handlers[method] = [handler]
            return handler

        return registerhandler

    def counttime(self):
        """
        This is not meantto be used outside the package!
        Use client.get_uptime() instead!
        """
        
        while True:
            self.uptime += 1
            time.sleep(1)

    async def call_event(self, type, *args, **kwargs):
        """
        This is not meantto be used outside the package!
        """

        Log.Debug(self, "Calling " + type + " event.")
        if type in self.handlers:
            for Method in self.handlers[type]:
                await Method(*args, **kwargs)

    async def set_status(self, state, activity=""):
        """
        This function is used to set the clients status and activity.
        Because of the way discord works, it might take a while to update status and activity.

        Parameters
        ----------

        Required: str State:
            "online" / "idle" / "dnd" / "invisible"
            something else wont work and results in an error.
        Currently required: str Activity:
            Example: "Counting sheep... zzz"
        """
        
        Log.Debug(self, "Setting status to {} and activity to {}...".format(state, activity))
        self.HTTPClient.Send_API(op.Status_update(state, activity))
        return

    async def get_guild(self, id):
        """
        This function returns a guild object.

        Parameters
        ----------
        Required: guild_id
            The guild id of the desired guild
        """
        
        guild = self.HTTPClient.get_guild(id)
        return guild

    async def get_channel(self, id):
        """
        This function returns a channel object.
        
        Parameters
        ----------
        Required: guild_id
            The channel id of the desired channel
        """

        channel = self.HTTPClient.get_channel(id)
        return channel

    async def get_guilds(self) -> list[Guild]:
        """
        This function will return a list with all guilds your client is in.
        """
        
        return self.guilds

    async def get_guild_count(self) -> int:
        """
        This function will return an integer that represents how many guilds the bot is currently in.
        """
        
        guilds = 0
        for guild in self.guilds:
            guilds+=1

        return guilds

    async def get_uptime(self) -> int:
        """
        This function will return an integer that represents how many seconds the bot has been online.
        """

        return self.uptime


import threading
import websocket
import json
import asyncio
import time
import requests

from .Log import Log
from .Guild import Guild
from .Message import Message

class HTTPClient:
    def __init__(self, client):
        #done
        self.client = client
        self.websocket = websocket.WebSocket()
        self.BaseUrl = "https://discord.com/api/v9/"
        self.headers = {"Authorization" : "Bot {}".format(self.client.Secret)}
        self.gatewayurl = "wss://gateway.discord.gg/?v=9&encoding=json"

    def connect(self):
        #done
        self.websocket.connect(self.gatewayurl)
        Log.Debug(self.client, "Connected to Websocket")
        
        self.Interval = json.loads(self.websocket.recv())["d"]["heartbeat_interval"]
        Log.Debug(self.client, "Received GateWay Info.")
        Log.Debug(self.client, "Heartbeat Interval set to " + str(self.Interval) + "ms")

        #send identifying OP
        self.websocket.send(op.identify(self.client.Token))

        #receiving info
        self.StartInfo = self.websocket.recv()
        Log.Debug(self.client, "Successfully identified GateWay and received Start info.")
        Log.Debug(self.client, "Collecting guild info in the process...")
        api = threading.Thread(target=self.Receive_API)
        api.start()
        self.HeartBeatThread = threading.Thread(target=self.HeartBeat)
        self.HeartBeatThread.start()
        return

    def HeartBeat(self):
        Log.Debug(self.client, "Sending initial HeartBeat")
        self.websocket.send(op.heartbeat())

        #Looping Heartbeat
        while True:
            time.sleep(self.Interval/1000-1)
            Log.Debug(self.client, "Sending HeartBeat")
            self.websocket.send(op.heartbeat())

    def Receive_API(self):
        #not done
        raw = self.websocket.recv()
        
        #Start new Receiving threads
        api = threading.Thread(target=self.Receive_API)
        api.start()
        
        signal = json.loads(raw)
        op = signal["op"]

        Log.Debug(self.client, "Received an OPCode " + str(op))
        #Dispatch
        if op == 0:
            name = signal["t"]
            data = signal["d"]

            if name == "GUILD_CREATE":
                self.client.guilds.append(Guild(data))

            if name == "MESSAGE_CREATE":
                message = Message(data, self)
                asyncio.run(self.client.call_event("on_message", message))

        #Reconnect
        elif op == 7:
            pass

        #Invalid Session
        elif op == 9:
            pass

        elif signal["op"] == 11:
            Log.Debug(self.client, "Received HeartBeat")

    def Send_API(self, dict):
        #done
        self.websocket.send(dict)
        return

    def send_message(self, ChannelID, strcontent):
        #done
        Url = self.BaseUrl + "channels/" + str(ChannelID) + "/messages"
        payload = json.dumps({"Content-Type": "application/json", "content":strcontent})
        headers = { "Authorization":"Bot {}".format(self.client.Token), "Content-Type":"application/json", }
        requests.post(Url, headers=headers, data=payload)
        return

    def get_guild(self, id):
        #Done
        Url = self.BaseUrl + "guilds/" + str(id)
        header = {"Authorization": "Bot {}".format(self.client.Token)}
        raw = requests.get(Url, headers=header)
        info = raw.json()
        return info

    def get_channel(self, id):
        #Done
        Url = self.BaseUrl + "channels/" + str(id)
        header = {"Authorization": "Bot {}".format(self.client.Token)}
        raw = requests.get(Url, headers=header)
        info = raw.json()
        return info

class op:
    def identify(Token):
        dict = '''
        {
            "op": 2,
            "d": {
                "token" : "''' + Token + '''",
                "intents":513,
                "properties":{
                    "$os":"linux",
                    "$browser":"my_library",
                    "$device":"my_library"
                    }
                }
            }
        '''
        return str(dict)
    
    def heartbeat():
        dict = '''
        {
            "op": 1,
            "d": null
        }
        '''
        return str(dict)

    def Status_update(status, activity):
        dict = '''
        {
            "op": 3,
            "d": {
                "since": null,
                "status": "'''+status+'''",
                "afk": false,
                "activities": [{
                    "name": "'''+activity+'''",
                    "type": 0
                }]
            }
        }
        '''
        return str(dict)


from .User import User
from .Channel import Channel

class Message:
    def __init__(self, data, HTMLClient):
        """
        This is the message class. It only stores variables at the time. 
        It can be replied by message.channel.send({content})
        """
        
        self.content = data["content"]
        self.id = data["id"]
        self.author = User(data["author"])
        self.channel = Channel(data["channel_id"], HTMLClient.get_channel(data["channel_id"]), HTMLClient)


class Channel:
    def __init__(self, ID, data, HTMLClient):
        self.Id = ID
        self.name = data["name"]
        self.guild_id = data["guild_id"]
        self.HTMLClient = HTMLClient

    async def send(self, content):
        """
        This method sends a message in this Channel.
        
        It requires one string parameter which represents the text sent as a discord message.
        """

        self.HTMLClient.send_message(self.Id, content)
        return

    async def get_guild(self):
        """
        This method returns the guild object of the guild this channel is in.

        It has no parameters.
        """

        guild = self.HTMLClient.get_guild(self.guild_id)
        return guild


from distutils.core import setup
import re

version = ''
with open('DiscAPI/__init__.py') as f:
  version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

setup(
  name = 'DiscAPI',         
  packages = ['DiscAPI'],   
  version = version,      
  license='MIT',        
  description = 'A simple to use Python Library to make all kinds of discord bot.',   
  author = "HiddenStorm",                   
  author_email = 'schuck2345@gmail.com',     
  url = 'https://github.com/Hiddenstorm/DiscAPI',  
  download_url = 'https://github.com/Hiddenstorm/DiscAPI/archive/refs/tags/v{}.tar.gz'.format(version),   
  keywords = ['Python', 'Discord', 'Simple'],   
  install_requires=[           
          'websocket-client'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # "3 - Alpha", "4 - Beta", "5 - Production/Stable" 
    'Intended Audience :: Developers',      
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',  
    'Programming Language :: Python :: 3.9',
  ],
)


from datetime import timedelta

class Log:
    def Debug(client, Content):
        if client.Log == True:
            time = client.uptime / 24 / 60 / 60
            print(bc.LOG+"[Log][{}] {}".format(str(timedelta(time)), Content) + bc.END)

    def Warning(Content):
        print(Content)
    
    def Error(Content):
        print(Content)
    
class bc:
    LOG = '\033[90m'
    WARNING = '\033[93m'
    ERROR = '\033[91m'
    END = '\033[0m'
